from dataclasses import dataclass
from typing import List

from ..models import Chapter


@dataclass
class Analysis:
    via_coins: List[Chapter]
    via_fastpass: List[Chapter]

    @staticmethod
    def empty():
        return Analysis(via_coins=[], via_fastpass=[])

    @property
    def coins_cost(self):
        return sum([c.cost for c in self.via_coins])

    @property
    def fastpass_cost(self):
        return len(self.via_fastpass)



from typing import List

from .analysis import Analysis
from ..models import Chapter


class IAnalyser:
    def analyse(self, chapters: List[Chapter]) -> Analysis:
        """
        :param chapters: chapters to be analysed. they must be locked
        :return: analysis object
        """
        raise NotImplementedError('method not overridden')



import functools
from copy import copy


class MethodDecorator:
    """
    __wrapped__: provides wrapped function
    __self__: provides wrapped functions class
    """

    def __init__(self, func):
        self.__self__ = None

        self.__wrapped__ = func
        functools.update_wrapper(self, func)

    def __call__(self, *args, **kwargs):
        if self.__self__ is None:
            raise ValueError

    def __get__(self, instance, owner):
        if instance is None:
            return self

        # create a bound copy
        bound = copy(self)
        bound.__self__ = instance

        # update __doc__ and similar attributes
        functools.update_wrapper(bound, self.__wrapped__)

        # add the bound instance to the object's dict so that
        # __get__ won't be called a 2nd time
        setattr(instance, self.__wrapped__.__name__, bound)

        return bound



from requests import Session


class HtmlApi:
    """
    provide html pages

    to be used primarily with BaseApi.html
    """

    def __init__(self, session: Session = None):
        """
        :param session: session to use for requests
        """
        if session is None:
            self.session = Session()
        else:
            self.session = session

    def profile(self):
        """
        :return: user profile html
        """
        return self.session.get(
            f'https://www.webnovel.com/profile/{self.session.cookies.get("uid")}?appId=10'
        ).content

    def vote(self):
        """
        :return: voting page html
        """
        return self.session.get(
            'https://www.webnovel.com/vote'
        ).content



import inspect

from .interface import MethodDecorator

URL_PARAM = 'url'


class redirect(MethodDecorator):
    def __call__(self, *args, **kwargs):
        super.__call__()

        url = None

        # check if url in kwargs
        if URL_PARAM in kwargs:
            url = kwargs[URL_PARAM]

        # check if url in args
        else:
            # get all args names of url
            # excluding the [self] arg
            args_names = inspect.getfullargspec(self.__wrapped__).args[1:]

            if URL_PARAM in args_names:
                # remove all parameters that are passed using kw
                # so args_names contains only those that haven't been passed as kw
                for kw in kwargs.keys():
                    args_names.remove(kw)

                index = args_names.index(URL_PARAM)
                try:
                    url = args[index]
                except IndexError:
                    pass

        if url is not None:
            self.__self__.driver.get(url)

        return self.__wrapped__(self.__self__, *args, **kwargs)



from dataclasses import dataclass

import requests
from bs4 import BeautifulSoup


@dataclass
class Novel:
    id: str = None
    title: str = None
    synopsis: str = None
    genre: str = None
    url: str = None
    cover_url: str = None

    author: str = None
    translator: str = None
    editor: str = None

    views: str = None
    rating: float = None
    review_count: int = None

    @staticmethod
    def from_url(url):
        """
        scrape novel information from webnovel

        :param url: url to novel
        :return: Novel object
        """
        response = requests.get(url)
        soup = BeautifulSoup(response.content, 'lxml')

        info_elems = soup.select('._mn > *')
        subinfo_elems = info_elems[1].select(':scope > *')
        writerinfo_elems = info_elems[2].select('p > *')

        novel = Novel()
        novel.id = int(url.split('/')[4])
        novel.title = info_elems[0].text[:-len(info_elems[0].find('small').text) - 1],
        novel.title = novel.title[0]

        novel.synopsis = soup.select_one("div[class*='j_synopsis'] > p").text
        novel.genre = subinfo_elems[0].text.strip(),
        novel.views = subinfo_elems[-1].text[:-6].strip(),
        novel.url = url[:]
        novel.cover_url = f'https://img.webnovel.com/bookcover/{novel.id}'

        try:
            # ratings are posted up to 5, they are converted to float and normalized to 1
            novel.rating = float(info_elems[3].find('strong').text) / 5.0,
            novel.rating = novel.rating[0]
        except ValueError:
            # no ratings
            novel.rating = None

        try:
            # stripped of all non numerals and converted to int
            novel.review_count = int(
                info_elems[3].find('small').text.strip('()')[:-8].replace(',', ''))
        except ValueError:
            # not enough reviews
            novel.review_count = None

        # writer info
        for i in range(round(len(writerinfo_elems) / 2)):
            label = writerinfo_elems[i * 2].text.strip(': ').lower()
            value = writerinfo_elems[i * 2 + 1].text

            setattr(novel, label, value)

        return novel



from typing import Optional, Iterable

import browser_cookie3
from requests.cookies import RequestsCookieJar


class Cookies(RequestsCookieJar):
    default_domains = (
        '.webnovel.com',
        'www.webnovel.com',
    )

    @classmethod
    def from_browser(cls, browser, domains: Optional[Iterable[str]] = None):
        """
        this function creates a Cookies object with the cookies from the given browser and the domains

        :param browser: browser from which to extract cookies
        :param domains: only cookies from these domains, when left empty uses `Cookies.default_domains`
        :raises TypeError: if the given browser is not supported
        :return: Cookies object containing the cookies
        """
        if domains is None:
            domains = cls.default_domains

        # check and see whether the expected browser is available
        try:
            func = getattr(browser_cookie3, browser)
        except AttributeError:
            raise TypeError(f'The provided browser ({browser}) is not supported')

        # create and fill the cookies jar
        cookiejar = Cookies()
        for domain in domains:
            for c in func(domain_name=domain):
                cookiejar.set(c.name, c.value, domain=c.domain, path=c.path)

        return cookiejar



from .interface import MethodDecorator
from ..exceptions import NotSignedInException


class require_signin(MethodDecorator):
    """
    signed in checker decorator class
    """

    def __call__(self, *args, **kwargs):
        super.__call__()

        if not self.__self__.is_signedin():
            raise NotSignedInException()

        return self.__wrapped__(self.__self__, *args, **kwargs)



from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from ..bot import WebnovelBot


class IHandler:
    bot: WebnovelBot
    driver: WebDriver

    def __init__(self, bot):
        self.bot = bot
        self.driver = bot.driver

        self.wait = WebDriverWait(self.driver, self.bot.timeout)

    def get(self, selector):
        return self.driver.find_element_by_css_selector(selector)

    def wait_for(self, selector):
        self.wait.until(
            EC.presence_of_element_located((By.CSS_SELECTOR, selector))
        )

    def wait_and_get(self, selector):
        self.wait.until(
            EC.presence_of_element_located((By.CSS_SELECTOR, selector))
        )

        return self.driver.find_element_by_css_selector(selector)



import json
import re
from typing import List, Union, Dict

from bs4 import BeautifulSoup
from requests.cookies import RequestsCookieJar
from selenium import webdriver
from selenium.common.exceptions import NoSuchElementException, TimeoutException, ElementClickInterceptedException
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from .analytic import IAnalyser, Analysis
from .api import ParsedApi, UnlockType
from .decorators import require_signin, redirect
from .exceptions import NotSignedInException, NotANovelUrlException, GuardException, CaptchaException
from .models import Profile, Novel, Chapter
from .tools import UrlTools

BASE_URL = 'https://www.webnovel.com'
EMAIL_RAW_URL = 'https://passport.webnovel.com/emaillogin.html'
EMAIL_LOGIN_URL = 'https://passport.webnovel.com/emaillogin.html?appid=900&areaid=1&auto=1&autotime=0&source=&ver=2' \
                  '&fromuid=0&target=iframe&option' \
                  '=&logintab=&popup=1&format=redirect '
GUARD_URL = 'https://passport.webnovel.com/guard.html'


class WebnovelBot:
    min_timeout = 10
    user_timeout = 600

    def __init__(self, driver: WebDriver = None, timeout=10):
        """
        :param driver: selenium web driver to use
        :param timeout: timeout for all class operations
        """
        if driver is None:
            self.driver = webdriver.Chrome()
        else:
            self.driver = driver

        self.timeout = timeout

    def create_api(self) -> ParsedApi:
        """
        :return: Api with cookies of the current selenium driver instance
        """
        return ParsedApi(self.driver.get_cookies())

    @property
    def novel_id(self):
        """
        get novel id from current url
        """
        url = self.driver.current_url
        if 'book' not in url:
            raise NotANovelUrlException

        return UrlTools.from_novel_url(url)

    @property
    def user_id(self):
        """
        :return: current user id
        """
        uid = self.driver.get_cookie('uid')

        # inline signing check
        if uid is None:
            raise NotSignedInException

        return int(uid['value'])

    def home(self):
        """
        load home
        """
        self.driver.get(BASE_URL)

    @require_signin
    def profile(self):
        """
        get profile information from popup

        :return: WebnovelProfile object
        """

        self._focus_profile()

        field_mapper = {
            'coins': 'Coins',
            'fastpass': 'Fast pass',
            'power_stone': 'Power Stones',
            'energy_stone': 'Energy Stones'
        }

        # redundancy check
        if list(field_mapper.keys()) != Profile.fields:
            raise ValueError('fields do not match')

        # generate dictionary using field_mapper as title value
        profile_data = {
            field: int(self.driver.find_element_by_css_selector(f"a[title='{value}'] > em").text)
            for field, value in field_mapper.items()
        }

        return Profile(id=self.user_id, **profile_data)

    def signin(self, email, password, manual=False):
        """
        signin to webnovel using :param email: and :param password:

        :param manual: setting it to true means that alogorithm can expect manual input when necessary.
        for example, during captcha or when signin is denied for reasons unknown

        :raises CaptchaException: if manual is false and captcha is required during signin
        :raises GuardException: if manual is false and signin process redirected to guard
        """
        # go to login path
        self.driver.get(EMAIL_LOGIN_URL)

        self.driver.find_element_by_class_name('loginEmail').send_keys(email)
        self.driver.find_element_by_class_name('loginPass').send_keys(password)

        signin_btn = self.driver.find_element_by_id('submit')
        signin_btn.click()

        wait = WebDriverWait(self.driver, self.timeout)
        wait_user = WebDriverWait(self.driver, self.user_timeout)

        # wait until redirected
        wait.until(lambda driver: (
                not self.driver.current_url.startswith(EMAIL_RAW_URL)
                or driver.find_elements_by_css_selector('#google-code-html iframe')  # checks if captcha appeared
        ))

        # this handles the cases where the captcha has appeared
        captchas = self.driver.find_elements_by_css_selector('#google-code-html iframe')
        if len(captchas) > 0:
            if not manual:
                raise CaptchaException

            # wait until user has solved the captcha
            wait_user.until(lambda driver: not self.driver.current_url.startswith(EMAIL_RAW_URL))

        # check if the redirected url is the guard, guard is shown when password is incorrect
        if self.driver.current_url.startswith(GUARD_URL):
            if not manual:
                raise GuardException

            wait_user.until(lambda driver: not self.driver.current_url.startswith(GUARD_URL))

        # wait till signin success ends
        WebDriverWait(self.driver, self.timeout).until(
            EC.presence_of_element_located((By.CSS_SELECTOR, "a[title='My Profile']"))
        )

        # wait for the preferences popup to load and click it away
        try:
            wait.until(
                EC.presence_of_element_located((By.CSS_SELECTOR, ".j_post_preference"))
            )

            pref_button = self.driver.find_element_by_class_name('j_post_preference')  # bt _m mw160 j_post_preference
            pref_button.click()
        except TimeoutException:
            pass

    def add_cookiejar(self, cookies: RequestsCookieJar):

        # selenium prevents adding cookies of domains that arent from the currently active url
        # this code makes sure we are in webnovel
        if not self.driver.current_url.startswith(BASE_URL):
            self.driver.get(BASE_URL)

        for cookie in cookies:
            self.driver.add_cookie({
                'name': cookie.name,
                'value': cookie.value,
                'domain': cookie.domain,
                'path': cookie.path,
            })

    @require_signin
    def signout(self):
        # click profile button to reveal the logout button
        self._focus_profile()

        # click and out
        signout_button = self.driver.find_element_by_css_selector("a[class*='j_logout']")
        signout_button.click()

        # wait till logged out
        WebDriverWait(self.driver, self.timeout).until(
            EC.presence_of_element_located((By.CSS_SELECTOR, '.login-btn'))
        )

    def is_signedin(self) -> bool:
        """
        checks for the existance of user id cookie

        :return: whether we are currently signed in
        """
        if self.driver.get_cookie('uid') is None:
            return False
        else:
            return True

    @require_signin
    def library(self, redirect: bool = True) -> List[Novel]:
        """
        :requires: to be signed in

        :return: all the novels in library
        """
        if redirect:
            self.driver.get(f'{BASE_URL}/library')

        # all novel containers
        novel_elements = self.driver.find_elements_by_css_selector('.lib-books > li')

        novels = []
        for element in novel_elements:
            link = element.find_element_by_css_selector('a')

            novel = Novel(
                title=link.text,
                url=link.get_attribute('href')
            )

            # this attribute does not exist in Novel class
            # ... it might be useful
            novel.update = '_update' in element.get_attribute('class')

            novels.append(novel)

        return novels

    @redirect
    def novel(self, url=None) -> Novel:
        """
        :param url: url to novel
        :return: Novel object
        """
        info_elems = self.driver.find_elements_by_css_selector('._mn > *')
        subinfo_elems = info_elems[1].find_elements_by_css_selector(':scope > *')
        writerinfo_elems = info_elems[2].find_elements_by_css_selector('p > *')

        novel = Novel()

        novel.id = self.driver.current_url.split('/')[4]
        novel.title = info_elems[0].text[:-len(info_elems[0].find_element_by_tag_name('small').text) - 1],
        novel.title = novel.title[0]

        novel.synopsis = self.driver.find_element_by_css_selector("div[class*='j_synopsis'] > p").text
        novel.genre = subinfo_elems[0].text,
        novel.views = subinfo_elems[-1].text[:-6],
        novel.url = self.driver.current_url
        novel.cover_url = f'https://img.webnovel.com/bookcover/{novel.id}'

        # ratings are posted up to 5, they are converted to float and normalized to 1
        novel.rating = float(info_elems[3].find_element_by_css_selector('strong').text) / 5.0,

        # stripped of all non numerals and converted to int
        novel.review_count = int(
            info_elems[3].find_element_by_css_selector('small').text.strip('()')[:-8].replace(',', ''))

        # writer info
        for i in range(round(len(writerinfo_elems) / 2)):
            label = writerinfo_elems[i * 2].text.strip(':').lower()
            value = writerinfo_elems[i * 2 + 1].text

            setattr(novel, label, value)

        return novel

    def table_of_contents(self) -> Dict[str, List[Chapter]]:
        """
        Must be used while a particular novel is loaded to driver

        :return: dict of volumes in order, where key is volume name and value the chapters
        """

        # load table of contents
        table_of_contents = self.driver.find_element_by_css_selector('a.j_show_contents')
        table_of_contents.click()

        # wait until table of contents loads
        WebDriverWait(self.driver, self.timeout).until(
            EC.presence_of_element_located((By.CLASS_NAME, 'volume-item'))
        )

        # using bs4 to parse
        # bs4 is faster than selenium selectors
        soup = BeautifulSoup(self.driver.page_source, 'html.parser')
        volume_elements = soup.find_all('div', {'class': 'volume-item'})
        volumes = {}
        for volume_element in volume_elements:
            volume_chapters = volume_element.find_all('li', {'class': 'g_col_6'})

            chapters = []
            for chapter in volume_chapters:
                no_element = chapter.select_one('a > i')

                chapter = Chapter(
                    title=chapter.select_one('a')['title'],
                    url=f"http:{chapter.find('a')['href'].strip()}",
                    locked=bool(chapter.select('a > svg'))
                )

                if no_element is not None:
                    chapter.no = int(no_element.text.strip())
                else:
                    chapter.no = 0

                chapter.id = chapter.chapter_id_from_url()

                chapters.append(chapter)

            _name = volume_element.find('h4').text
            volume_name = re.sub(r'\n +', '', _name).replace(':', ': ')
            volumes[volume_name] = chapters

        return volumes

    @redirect
    def chapter(self, url=None, is_locked=False) -> Chapter:
        """
        :param url: url to chapter
        :param is_locked: whether the chapter is locked
        :return: Chapter object
        """
        # wait till chapter loads
        wait = WebDriverWait(self.driver, self.timeout)
        wait.until(
            EC.presence_of_element_located((By.CLASS_NAME, 'cha-tit'))
        )

        # if locked, ensure lock element loads
        # lock element may take some time to load
        if is_locked:
            wait.until(
                EC.presence_of_element_located((By.CSS_SELECTOR, '.j_locked_chap'))
            )

        chapter = Chapter(
            no=int(self.driver.find_element_by_css_selector('.j_chapIdx').text.strip('Chapter ')[:-1]),
            url=self.driver.current_url,
            title=self.driver.find_element_by_css_selector('.cha-tit').find_element_by_tag_name('h3').text,
            locked=self.is_chapter_locked(),
        )

        if chapter.locked:
            if self.is_signedin():
                unlock_params = self.driver.find_element_by_css_selector(
                    "div[class*='lock-group j_lock_btns '] > a[class*='j_unlockChapter']"
                ).get_attribute('data-unlock-params')
            else:
                unlock_params = self.driver.find_element_by_css_selector(
                    "div[class*='lock-group j_lock_btns '] > a[class*='_bt_unlock']"
                ).get_attribute('data-unlock-params')

            chapter.cost = json.loads(unlock_params)['chapterPrice']
        else:
            chapter.paragraphs = [wrapper.text
                                  for wrapper in self.driver.find_elements_by_css_selector('.cha-paragraph > span > p')]

        return chapter

    @redirect
    def is_chapter_locked(self, url=None):
        return bool(self.driver.find_elements_by_css_selector('.j_locked_chap'))

    def unlock_chapter(self, url=None, coins=False, fastpass=False, unlock_delay=2):
        """
        unlocks the chapter using either coins or fastpass

        either one of [coins] and [fastpass] must be true
        if both options are available, [coins] are given priority

        :required: to be signed in

        :param url: url to chapter
        :param coins: use coins to unlock
        :param fastpass: use fastpass to unlock
        :param unlock_delay: delay between mouse moving to unlock button to clicking.
                             this is introduced as unlock can fail due to some latency issues
        """
        # error testing parameters
        if not coins and not fastpass:
            raise ValueError('Choose atleast one from "coins" or "fastpass"')

        # manual [url] and [signin] check
        # as [coins] and [fastpass] check needs priority
        if url is not None:
            self.driver.get(url)

        # sign in check
        if not self.is_signedin():
            raise NotSignedInException()

        # wait till lock element loads
        WebDriverWait(self.driver, self.timeout).until(
            EC.presence_of_element_located((By.CSS_SELECTOR, '.j_locked_chap'))
        )

        # check which options are available [coins, fastpass]
        # coins
        try:
            can_coin = True
            self.driver.find_element_by_css_selector("div[class='lock-group j_lock_btns ']")
        except NoSuchElementException:
            can_coin = False

        # fastpass
        try:
            can_fastpass = True
            self.driver.find_element_by_css_selector("div[class^='lock-foot'] > div:nth-child(2) > a")
        except NoSuchElementException:
            can_fastpass = False

        def unlock(button):
            """
            :param button: button to click to unlock
            """
            ActionChains(self.driver).move_to_element(button).pause(unlock_delay).click().perform()

        if coins and can_coin:
            coin_button = self.driver.find_element_by_css_selector(
                "div[class='lock-group j_lock_btns '] > a[class*='j_unlockChapter']")
            unlock(coin_button)

        elif fastpass and can_fastpass:
            fastpass_button = self.driver.find_element_by_css_selector("div[class^='lock-foot'] > div:nth-child(2) > a")
            unlock(fastpass_button)

    def batch_unlock(self, analysis, selenium=False):
        """
        unlocks all the chapters in analysis according to assortment
        Uses api to do post requests by default

        :require: to be signed in

        :param selenium: [Outdated] whether to use selenium to unlock, default is [False]
        :param analysis: object to denote which to unlock in which
        :return: None
        """
        if selenium:
            # [outdated]
            # signed in check not in method
            # as mainly consisting of [unlock_chapter] which has signin check
            for c in analysis.via_coins:
                self.unlock_chapter(c.url, coins=True)
            for c in analysis.via_fastpass:
                self.unlock_chapter(c.url, fastpass=True)

        else:
            api = self.create_api()
            for c in analysis.via_coins:
                api.unlock(c.novel_id_from_url(), c, UnlockType.coins)
            for c in analysis.via_fastpass:
                api.unlock(c.novel_id_from_url(), c, UnlockType.fastpass)

    @redirect
    def batch_analyze(self, analyser: IAnalyser, url=None) -> Analysis:
        """
        batch unlocks chapters in ascending order

        :require: to be signed in

        :param analyser: analyser for chapters to unlock
        :param url: url to novel
        :return: list of unlocked chapters
        """
        # get all locked chapters
        locked_chapters = [chapter for chapters in self.table_of_contents().values() for chapter in chapters]

        return analyser.analyse(locked_chapters)

    @require_signin
    def energy_vote(self, votes: Union[int, List[int]], lead='male', redirect=True):
        """
        votes for the selected indexes of the queue of stories to be released

        :requires: to be signed in

        :param votes: number of votes to cast or indexes of stories in queue to vote
        :param lead: either 'male' or 'female' options, to specify which list to select
        :param redirect: whether to load the vote page again
        """

        # to energy_vote screen
        if redirect:
            self.driver.get(f'{BASE_URL}/vote')

        # set indexes to be unlocked
        indexes = None
        if type(votes) == list:
            indexes = votes
        elif type(votes) == int:
            indexes = range(votes)

        lead_list = self.driver.find_element_by_css_selector(f'#List{lead.capitalize()}')
        vote_buttons = lead_list.find_elements_by_css_selector('.j_canVote._voteBtn')

        # this checks whether user can vote / has energy stones
        if not vote_buttons:
            return

        for i in indexes:
            vote_buttons[i].click()

    @require_signin
    def claim_tasks(self):
        """
        collects all completed task rewards

        :requires: to be signed in
        """
        # click profile button to reveal show claim_tasks button
        self._focus_profile()

        # click to show claim_tasks
        self.driver.execute_script("document.querySelector('.j_show_task_mod').click()")

        # wait till tasks are loaded
        wait = WebDriverWait(self.driver, timeout=self.timeout)
        wait.until(
            EC.presence_of_element_located((By.CSS_SELECTOR, '.task-list-item'))
        )

        # claim all rewards
        for claim in self.driver.find_elements_by_css_selector('.j_claim_task'):
            claim.click()

        # exit
        self.driver.execute_script("document.querySelector('#taskMod ._close').click()")

        # wait till popup closes
        wait.until_not(
            EC.presence_of_element_located((By.CSS_SELECTOR, "div#taskMod._on"))
        )

    @require_signin
    def power_vote(self, url: str = None, repeat: int = 1):
        """
        energy_vote with power stone to [url]

        :requires: to be signed in

        :param url: url of novel to power up, if this is not specified tries to energy_vote in current page

        :param repeat: number of times to press button.
        :default repeat: 1
        """
        if url is not None:
            self.driver.get(url)
        elif re.fullmatch(r'https://.+?/book/.+?(/|$)', self.driver.current_url):
            raise ValueError('currently loaded url is not a valid novel url')

        WebDriverWait(self.driver, self.timeout).until(
            EC.presence_of_element_located((By.CSS_SELECTOR, '.j_vote_power'))
        )

        # get power stone energy_vote button
        # exception is thrown when user has no power stones
        try:
            power_button = self.driver.find_element_by_css_selector(".j_vote_power._on")
        except NoSuchElementException:
            return

        # using an action chain to delay each subsequent click since
        # webnovel button animation blocks clicks
        chain = ActionChains(self.driver)
        chain.move_to_element(power_button)

        for _ in range(repeat):
            chain.click().pause(1)

        chain.perform()

    def close(self):
        self.driver.close()

    def _focus_profile(self):
        """
        moves mouse over to profile button

        :return: profile button element
        """
        profile_button = self.driver.find_element_by_css_selector("div[class^='g_user'][class*='g_dropdown']")

        # even though the exception is thrown (with good reason) it seems to open profile
        try:
            profile_button.click()
        except ElementClickInterceptedException:
            pass

        return profile_button



from typing import Tuple


class UrlTools:
    """
    url conversion tool to and from [novel_id], [chapter_id], and [profile_id]
    """
    base_url = 'https://www.webnovel.com/book'

    @staticmethod
    def to_novel_url(novel_id):
        """
        :return: novel url
        """
        return f'{UrlTools.base_url}/{novel_id}'

    @staticmethod
    def from_novel_url(novel_url):
        """
        :return: url of novel
        """
        try:
            return int(novel_url.split('/')[4])
        except ValueError:
            return int(novel_url.split('_')[-1])

    @staticmethod
    def to_chapter_url(novel_id, chapter_id):
        """
        :return: chapter url
        """
        return f'{UrlTools.base_url}/{novel_id}/{chapter_id}'

    @staticmethod
    def from_chapter_url(chapter_url) -> Tuple[int, int]:
        """
        :return: novel_id, chapter_id
        """
        pieces = chapter_url.split('/')
        try:
            return int(pieces[4]), int(pieces[5])
        except ValueError:
            return int(pieces[4].split('_')[-1]), int(pieces[5].split('_')[-1])

    @staticmethod
    def to_profile_url(profile_id):
        """
        :return: profile url
        """
        return f'https://www.webnovel.com/profile/{profile_id}?appId=10'

    @staticmethod
    def from_profile_url(profile_url):
        """
        :return: profile id
        """
        return int(profile_url.split('/')[4].split('?')[0])



class NotSignedInException(Exception):
    """
    Raised when user not signed in to webnovel
    """
    pass


class NotANovelUrlException(Exception):
    """
    Raised when the link is not a novel url
    """
    pass


class ApiError(Exception):
    pass


class GuardException(Exception):
    pass


class CaptchaException(Exception):
    pass


class ValidationError(Exception):
    pass



import json
from dataclasses import dataclass
from typing import List

import requests
from bs4 import BeautifulSoup

from ..decorators import deprecated


@dataclass
class Chapter:
    no: int = None
    id: int = None
    url: str = None
    title: str = None
    paragraphs: List[str] = None
    locked: bool = None
    cost: int = None
    type: int = None

    @deprecated('use Webnovel.api.ParsedApi.chapter instead')
    def load(self):
        data = requests.get(self.url)

        soup = BeautifulSoup(data.content, 'html.parser')
        locked_content = soup.find('div', {'class': 'j_locked_chap'})

        self.id = self.chapter_id_from_url()

        self.no = int(soup.find('span', {'class': 'j_chapIdx'}).text.strip('Chapter ')[:-1])
        self.title = soup.find('div', {'class': 'cha-tit'}).find('h3').text
        self.locked = locked_content is not None

        if self.locked:
            params = json.loads(locked_content.find('a', {'class': '_bt_unlock'})['data-unlock-params'])
            self.cost = params['chapterPrice']

        return self

    def chapter_id_from_url(self) -> int:
        return int(self.url.split('/')[5])

    def novel_id_from_url(self) -> int:
        return int(self.url.split('/')[4])



from functools import wraps


def chainable(func):
    """
    enables chaining of methods

    this makes the return of the function to be the class or `self`

    example usage:

        @chainable
        def a_class_method(self):
            ...

    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        func(*args, **kwargs)

        # when used with a method, the first arg is always the class
        return args[0]

    return wrapper



from dataclasses import dataclass

from bs4 import BeautifulSoup


@dataclass
class Profile:
    id: int = None
    coins: int = None
    fastpass: int = None
    power_stone: int = None
    energy_stone: int = None

    fields = ['coins', 'fastpass', 'power_stone', 'energy_stone']

    @staticmethod
    def from_html(html: str, user_id=None):
        """
        get profile info using user id

        [html] can be obtained through, BaseApi.html.profile with cookies.
        [user_id] can be obtained through WebnovelBot.user_id.

        :param html: profile page
        :param user_id: id of user profile
        :return: profile of user
        """
        soup = BeautifulSoup(html, 'lxml')

        profile = Profile(id=user_id)

        group = soup.select("div[class='fl'] > a.dib.mr32, div[class='fl'] > a.dib.mr16")

        profile.coins = int(group[0].text)
        profile.fastpass = int(group[1].text)
        profile.power_stone = int(group[2].text)
        profile.energy_stone = int(group[3].text)

        return profile



from selenium.common.exceptions import NoSuchElementException

from .handler import IHandler
from ..bot import GUARD_URL
from ..decorators import chainable


class GuardHandler(IHandler):
    def __init__(self, bot):
        """
        :param bot: webnovel bot
        :raises ValueError: if current url is not a guard url
        """
        super(GuardHandler, self).__init__(bot)

        if not self.driver.current_url.startswith(GUARD_URL):
            raise ValueError('current page is not a guard')

        # make sure its fully loaded
        self.wait_for('.codeInfo > input')

        self.input_element = self.get('.codeInfo > input')
        self.confirm_buttom = self.get('#checkTrust')
        self.resend_button = self.get('#resTrustEmail')
        self.back_button = self.get('.m-main-hd a')

    @chainable
    def input(self, code: str):
        """
        inputs code into authentication input field

        :param code: authentication code
        """
        self.input_element.send_keys(code)

    @chainable
    def confirm(self):
        """ press confirm button """
        self.confirm_buttom.click()

    @chainable
    def resend(self):
        """ press resend email button """
        self.resend_button.click()

    def back(self):
        """ press back button """
        self.back_button.click()

    def wait_until_confirmed(self):
        self.wait.until(lambda driver: (
                not self.driver.current_url.startswith(GUARD_URL)
                or self.driver.find_elements_by_css_selector('.error_tip._on')
        ))

        try:
            self.get('.error_tip._on')
        except NoSuchElementException:
            return True
        else:
            return False



from typing import List, Callable

from . import Analysis
from .interface import IAnalyser
from ..api import ParsedApi
from ..models import Chapter, Profile, Novel


class HardLine(IAnalyser):
    """
    Calculate while abiding by the rules:
    - chapter cost less than or equal to [coins_line] would be selected as coins
    - chapter cost higher than or equal to [fastpass_line] would be select as fastpass
    """

    def __init__(self, novel: Novel, profile: Profile, coins_line: int = None, fastpass_line: int = None,
                 on_load: Callable[[Chapter], None] = None):
        """
        [fastpass_line] and [coins_line] are used to determine which is selected for which

        :param novel: attribute id must not be null
        :param profile: webnovel profile, require [coins] and [fastpass]
        :param coins_line: chapters with cost less that or equal are selected for coins
        :param fastpass_line: chapters with cost greater than or equal are selected for fastpass
        :raises ValueError: if [fastpass_line] is less than [coins_line] or if both lines are None
        """
        self.novel = novel
        self.profile = profile

        if coins_line is not None and fastpass_line is not None:
            if coins_line >= fastpass_line:
                raise ValueError('[coins_line] must be less than [fastpass_line]')
        elif coins_line is None and fastpass_line is None:
            raise ValueError('either [coins_line] or [fastpass_line] must not be "None"')

        self.fastpass_line = fastpass_line
        self.coins_line = coins_line

        if on_load is None:
            self.on_load = lambda c: None
        else:
            self.on_load = on_load

        self._api = ParsedApi()

    def analyse(self, chapters: List[Chapter]) -> Analysis:

        coins = self.profile.coins
        fastpass = self.profile.fastpass

        # flags
        coins_maxed = self.coins_line is None or coins == 0
        fastpass_maxed = self.fastpass_line is None or fastpass == 0

        analysis = Analysis.empty()
        for chapter in chapters:

            # making sure chapter has cost attribute populated
            # if not requesting for chapter cost
            if chapter.cost is None:
                chapter = self._api.chapter(self.novel.id, chapter.id)

            self.on_load(chapter)

            coins_cost = analysis.coins_cost
            fastpass_cost = analysis.fastpass_cost

            # distribute
            if not coins_maxed:
                # check whether adding the chapter will exceed coins balance
                if coins_cost + chapter.cost <= coins and chapter.cost <= self.coins_line:
                    analysis.via_coins.append(chapter)
                    coins_cost += chapter.cost

            elif not fastpass_maxed and chapter.cost >= self.fastpass_line:
                analysis.via_fastpass.append(chapter)

                fastpass_cost += 1
                fastpass_maxed = fastpass == fastpass_cost

            if coins_maxed and fastpass_maxed:
                break

        return analysis



from typing import List, Callable

from .analysis import Analysis
from .interface import IAnalyser
from ..api import ParsedApi
from ..models import Profile, Chapter, Novel


class ForwardCrawl(IAnalyser):
    """
    Calculate while abiding by the rules:
     - all chapters must be unlocked continuously
     - cheaper chapters with coins
     - expensive chapters with fastpass
     - chapters unlocked using coins must not exceed [maximum_cost] individually

    if [maximum_cost] is less than 0 it is considered as being infinite
    """

    def __init__(self, novel: Novel, profile: Profile, maximum_cost=-1, on_load: Callable[[Chapter], None] = None):
        """
        :param novel: attribute id must not be null
        :param profile: webnovel profile, require [coins] and [fastpass]
        :param maximum_cost: maximum coins to spend a single chapter, maximum
        :param on_load: call when chapter is loaded
        """
        self.novel = novel
        self.profile = profile
        self.maximum_cost = maximum_cost

        if on_load is None:
            self.on_load = lambda c: None
        else:
            self.on_load = on_load

        self._api = ParsedApi()

    def analyse(self, chapters: List[Chapter]) -> Analysis:

        coins = self.profile.coins
        fastpass = self.profile.fastpass

        explored = []

        # skip chapters count of fastpass
        # each fastpass can unlock a chapter
        # so length of fastpass can be unlocked
        # if count is 3, start from index 3, item 4
        i = fastpass

        # iterate through fastpass count
        for c in chapters[:i]:
            chapter = self._api.chapter(self.novel.id, c.id)
            self.on_load(chapter)

            explored.append(chapter)

        while True:
            # add new chapter
            chapter = self._api.chapter(self.novel.id, chapters[i].id)
            self.on_load(chapter)

            explored.append(chapter)

            # sort according to cost
            explored.sort(key=lambda ch: ch.cost)

            # split such that the ones that cost most are assigned to fastpass
            trial_coins, trial_fastpass = self._split(explored, fastpass)

            # check if valid solution
            total_trial_coins: int = sum([c.cost for c in trial_coins])

            # if maximum cost is positive check chapters
            # else maximum cost is regarded as being infinite.
            overshooted = any([c.cost > self.maximum_cost for c in trial_coins]) if self.maximum_cost >= 0 else False

            # if not valid option
            # previous solution must possible
            # adding the last chapter caused solution to overshoot
            # hence last solution must have maximum unlockable
            if overshooted or coins < total_trial_coins:
                # remove last added to get solution
                explored.remove(chapter)

                # sort according to cost
                explored.sort(key=lambda ch: ch.cost)

                # split such that the ones that cost most are assigned to fastpass
                trial_coins, trial_fastpass = self._split(explored, fastpass)

                # return solution
                return Analysis(via_coins=trial_coins, via_fastpass=trial_fastpass)

            i += 1

    def _split(self, chapters: List, fastpass: int):
        """
        split chapters so that most expensive will be to fastpass

        :param chapters: sorted chapters to split
        :param fastpass: amount of fastpass
        :return: coins, fastpass
        """
        if fastpass == 0:
            return chapters, []
        else:
            return chapters[:-fastpass], chapters[-fastpass:]



from typing import List, Dict

from bs4 import BeautifulSoup

from .base import BaseApi
from ..models import Chapter


class UnlockType:
    coins = 3
    fastpass = 5


class ParsedApi(BaseApi):
    def toc(self, novel_id) -> Dict[str, List[Chapter]]:
        response = super().toc(novel_id)

        volume_items = response['data']['volumeItems']

        volumes = {}
        for volume_item in volume_items:
            chapters = [
                Chapter(
                    no=item['index'],
                    id=item['id'],
                    url=f'https://www.webnovel.com/book/{novel_id}/{item["id"]}',
                    title=item['name'],
                    locked=not int(item['isAuth']),
                ) for item in volume_item['chapterItems']
            ]

            _name = volume_item['name']
            volume_name = f'Volume {volume_item["index"]}' + (f': {_name}' if _name else '')
            volumes[volume_name] = chapters

        return volumes

    def chapter(self, novel_id, chapter_id) -> Chapter:
        response = super().chapter(novel_id, chapter_id)

        data = response['data']['chapterInfo']

        return Chapter(
            no=data['chapterIndex'],
            id=data['chapterId'],
            title=data['chapterName'],
            url=f'https://www.webnovel.com/book/{novel_id}/{chapter_id}',
            paragraphs=[para['content'] for para in data['contents']],
            cost=data['price'],
            locked=not int(data['isAuth']),
            type=int(data['chapterLevel']),
        )

    def unlock(self, novel_id: int, chapter: Chapter, unlock_type: int) -> Chapter:
        """
        Unlocks chapters provided with options specified

        :param novel_id: corresponding novel
        :param chapter: chapter to unlock
        :param unlock_type: use UnlockType attributes to get correct ints
        :return: unlocked chapters
        """

        # check and convert to viable form data
        form_data = [{
            'chapterPrice': chapter.cost if unlock_type == UnlockType.coins else 1,
            'chapterId': str(chapter.id),
            'chapterType': 2  # no idea why its 2, but it seems to work
        }]

        response = super().unlock(novel_id, form_data, unlock_type)

        data = response['data']

        # update paragraph data
        chapter.paragraphs = [BeautifulSoup(para['content'], 'lxml').text for para in data['contents']]

        return chapter



from setuptools import setup, find_packages

import webnovel

with open('README.md', 'r') as f:
    long_description = f.read()

with open('requirements.txt', 'r') as f:
    requirements = f.readlines()

setup(
    name='webnovelbot',
    version=webnovel.__version__,
    author="Schicksal",
    description="webnovel scraper using selenium",
    author_email='mhaisham79@gmail.com',
    long_description=long_description,
    long_description_content_type='text/markdown',

    install_requires=requirements,

    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',
        "Operating System :: OS Independent",
    ],
    license="MIT license",
    keywords='webnovel novel bot api',

    url='https://github.com/mHaisham/webnovelbot',
    project_urls={
        'Source code': 'https://github.com/mHaisham/webnovelbot'
    },
    packages=find_packages(),
    python_requires='>=3.6'
)



from typing import List, Callable

from . import Analysis
from .interface import IAnalyser
from ..api import ParsedApi
from ..models import Chapter, Novel, Profile


class Efficient(IAnalyser):
    """
    Calculate while abiding by the rules:
     - cheaper chapters with coins
     - expensive chapters with fastpass
     - chapters unlocked using coins must not exceed [maximum_cost] individually

    if [maximum_cost] is less than 0 it is considered as being infinite
    """

    def __init__(self, novel: Novel, profile: Profile, maximum_cost=-1, on_load: Callable[[Chapter], None] = None):
        """
        :param novel: attribute id must not be null
        :param profile: webnovel profile, require [coins] and [fastpass]
        :param maximum_cost: maximum coins to spend a single chapter, maximum
        :param on_load: call when chapter is loaded
        """
        self.novel = novel
        self.profile = profile
        self.maximum_cost = maximum_cost

        if on_load is None:
            self.on_load = lambda c: None
        else:
            self.on_load = on_load

        self._api = ParsedApi()

    def analyse(self, chapters: List[Chapter]) -> Analysis:

        coins = self.profile.coins
        fastpass = self.profile.fastpass

        if coins <= 0 and fastpass <= 0:
            return Analysis.empty()

        # get cost
        for i in range(len(chapters)):
            if chapters[i].cost is None:
                chapters[i] = self._api.chapter(self.novel.id, chapters[i].id)
            self.on_load(chapters[i])

        # sort according to cost
        chapters.sort(key=lambda c: c.cost)

        if fastpass == 0:
            not_chosen = chapters
            via_fastpass = []
        else:
            not_chosen, via_fastpass = chapters[:-fastpass], chapters[-fastpass:]

        # get smallest possible coins cost to fit profile
        via_coins: List[Chapter] = []
        for chapter in not_chosen:
            total_cost = sum([c.cost for c in via_coins])

            if total_cost + chapter.cost > coins:
                break

            # maximum cost less than 0, means maximum cost is ignored
            # ending search as subsequent chapters would have higher or equal cost as list is sorted
            if 0 <= self.maximum_cost < chapter.cost:
                break

            via_coins.append(chapter)

        return Analysis(via_coins=via_coins, via_fastpass=via_fastpass)



import json
from typing import Dict, List, Union

import requests
from requests.cookies import RequestsCookieJar

from .cookie import BlockAll
from .html import HtmlApi
from ..exceptions import ApiError


class BaseApi:
    """
    Creates direct requests to data api rather than to load web page
    """

    def __init__(self, cookies: Union[RequestsCookieJar, List[dict], None] = None):
        # as cookies can lead to a rejected response
        # they are all blocked
        self.session = requests.Session()
        self.has_cookies = bool(cookies)

        # set cookies
        if type(cookies) == RequestsCookieJar:
            self.session.cookies = cookies
        elif type(cookies) == list:
            for cookie in cookies:
                cookie = {key: value for key, value in cookie.items() if key not in ['httpOnly', 'expiry', 'sameSite']}
                m_cookie = requests.cookies.create_cookie(**cookie)
                self.session.cookies.set_cookie(m_cookie)
        elif cookies is None:
            self.session.cookies.set_policy(BlockAll())
        else:
            raise TypeError("'cookies' was of unrecognized type; must be (RequestsCookieJar, List[dict cookie], None)")

        # html api
        self.html = HtmlApi(self.session)

    def chapter(self, novel_id: int, chapter_id: int) -> Dict:
        response = self.session.get(
            'https://www.webnovel.com/go/pcm/chapter/getContent',
            params={
                # '_csrfToken': self.session.cookies.get('_csrfToken') if self.has_cookies else '',
                'bookId': novel_id,
                'chapterId': chapter_id
            }
        )

        return self.validate(response)

    def toc(self, novel_id: int) -> Dict:
        response = self.session.get(
            'https://www.webnovel.com/apiajax/chapter/GetChapterList',
            params={
                '_csrfToken': self.session.cookies.get('_csrfToken') if self.has_cookies else '',
                'bookId': novel_id,
            }
        )

        return self.validate(response)

    def unlock(self, novel_id: int, chapters: List[Dict], unlock_type: int):
        """
        Unlocks chapters provided with method specified

        Requires cookies

        :param novel_id: id of novel
        :param chapters: list of dict objects containing the following keys,
         [chapterPrice: int],
         [chapterId: str],
         [chapterType: int]

        :param unlock_type: 3 to unlock with coins | 5 to unlock with fastpass
        :return:
        """

        response = self.session.post(
            'https://www.webnovel.com/apiajax/SpiritStone/useSSAjax',
            data={
                '_csrfToken': self.session.cookies.get('_csrfToken') if self.has_cookies else '',
                'bookId': novel_id,
                'chapters': json.dumps(chapters),
                'unlockType': unlock_type
            }
        )

        return self.validate(response)

    def power_vote(self, novel_id):
        """
        Applies single power stone vote to novel

        Requires cookies

        :param novel_id: novel to vote
        :return: Response
        """
        response = self.session.post(
            'https://www.webnovel.com/apiajax/powerStone/vote',
            data={
                '_csrfToken': self.session.cookies.get('_csrfToken') if self.has_cookies else '',
                'bookId': novel_id,
                'novelType': 0
            }
        )

        return self.validate(response)

    def energy_vote(self, novel_id):
        """
        Applies single energy stone vote to translation novel release queue

        Requires cookies

        :param novel_id: novel to vote
        :return: Response
        """
        response = self.session.post(
            'https://www.webnovel.com/apiajax/translationVote/vote',
            data={
                '_csrfToken': self.session.cookies.get('_csrfToken') if self.has_cookies else '',
                'bookId': novel_id
            }
        )

        return self.validate(response)

    def validate(self, response) -> Dict:
        parsed = json.loads(response.content)
        if parsed['code'] != 0:
            raise ApiError(parsed['msg'])

        return parsed
